<?php

/**
 * @file
 * Batch processing API for populate Media field of the entity with image items.
 */

use Drupal\Core\Entity\EntityInterface;

/**
 * Populate Media field of the entity with image items.
 *
 * Create Media entities based on existing image fields.
 * Populate the Media field, that was created, with the Media entities.
 *
 * @param \Drupal\Core\Entity\EntityInterface $entity
 *   The entity that has not empty Image field.
 * @param string $image_field_name
 *   The name of the Image field.
 * @param string $media_field_name
 *   The name of the created Media field.
 * @param array $context
 *   Batch context.
 */
function image_field_to_media_populate_media_field(EntityInterface $entity, $image_field_name, $media_field_name, array &$context) {
  $images = $entity->get($image_field_name)->getValue();
  $entity->set($media_field_name, NULL);

  foreach ($images as $image) {
    $media = image_field_to_media_get_media_entity($image);
    $entity->get($media_field_name)->appendItem($media);
  }

  $entity->save();
  $context['results']['updated_entities']++;
}

/**
 * Return or new or existed Media entity.
 *
 * If a Media entity with such image already exists, then return this entity.
 * To implement this we store sha1 hashes of image files after creation of
 * new Media entities.
 *
 * @param array $image
 *   The array that represents one item of the Image field.
 *
 * @return \Drupal\media\MediaInterface
 *   The media Entity with the image.
 */
function image_field_to_media_get_media_entity(array $image) {
  // Get stored hashes of image files.
  $hashes = \Drupal::state()->get('image_field_to_media.hashes_of_image_files');

  // Get hash of the current image file.
  $fid = $image['target_id'];
  $fileUri = \Drupal::entityTypeManager()
    ->getStorage('file')
    ->load($fid)
    ->getFileUri();

  $hash = sha1_file($fileUri);

  // Check if the Media entity with such image already exists.
  // Compare hash of the image file with hashes that was stored previously.
  // The structure of the $hashes array: ['media id' => 'hash of image file']
  // 'media id' is id of the Media entity that contains the image file that
  // was hashed.
  $mid = array_search($hash, $hashes);
  if ($mid) {
    // Load the Media entity that has the current image and return it.
    $media = \Drupal::entityTypeManager()
      ->getStorage('media')
      ->load($mid);

    return $media;
  }

  // Create new Media entity with the current image.
  $media = \Drupal::entityTypeManager()
    ->getStorage('media')
    ->create([
      'bundle' => 'image',
      'uid' => 1,
    ]);

  $media->set('field_media_image', $image);
  $media->save();

  // Store the media id and hash of the image file.
  $hashes[$media->id()] = $hash;
  \Drupal::state()->set('image_field_to_media.hashes_of_image_files', $hashes);

  return $media;
}

/**
 * Batch finish callback.
 */
function image_field_to_media_batch_finished($success, $results, $operations) {
  if ($success) {
    $entities_number = $results['updated_entities'];
    $message = t('Cloned successfully! Updated entities: @number.', ['@number' => $entities_number]);
    \Drupal::messenger()->addStatus($message);
  }
  else {
    \Drupal::messenger()->addError(t('Finished with an error.'));
  }
}
